use std::{fs::File, io::Write, sync::{Arc, Mutex}};

use board::Board;
use ruleset::Ruleset;
use seen::Seen;

use crate::dumb_dealer::Deal;

mod board;
mod ruleset;
mod pico_rng;
mod seen;
mod dumb_dealer;
mod zobrist;


fn main() {
    /*
    let filename = "level_1.txt";
    let ruleset = Ruleset { 
        n_slots: 5, 
        n_suits: 1, 
        n_cards_per_suit: 9, 
        n_arcana: 0 
    };
    */
    /*
    let filename = "level_2.txt";
    let ruleset = Ruleset { 
        n_slots: 5, 
        n_suits: 2, 
        n_cards_per_suit: 9, 
        n_arcana: 0 
    };
    */
    /* 
    let filename = "level_3.txt";
    let ruleset = Ruleset { 
        n_slots: 7, 
        n_suits: 2, 
        n_cards_per_suit: 9, 
        n_arcana: 8 
    };
    */
    /*
    let filename = "level_4.txt";
    let ruleset = Ruleset { 
        n_slots: 9, 
        n_suits: 3, 
        n_cards_per_suit: 9, 
        n_arcana: 16 
    }; 
    */
    /*
    let filename = "level_5.txt";
    let ruleset = Ruleset { 
        n_slots: 9, 
        n_suits: 3, 
        n_cards_per_suit: 11, 
        n_arcana: 18 
    }; 
    */
    /*
    let filename = "fortunes_foundation.txt";
    let ruleset = Ruleset { 
        n_slots: 11, 
        n_suits: 4, 
        n_cards_per_suit: 13, 
        n_arcana: 22 
    };
    */

    let filename = "level_7.txt";
    let ruleset = Ruleset { 
        n_slots: 11, 
        n_suits: 5, 
        n_cards_per_suit: 10, 
        n_arcana: 25 
    };

    let setup = ruleset.compile().expect("compilation should succeed");

    const THREADS: usize = 28;
    let work_source: WorkSource = WorkSource::new(0);
    let winnable_seeds_file = Arc::new(Mutex::new(File::create(filename).expect("should be able to create")));

    let mut threads = Vec::new();
    for _ in 0..THREADS {
        let setup2 = setup.clone();
        let source = work_source.clone();
        let wsf = winnable_seeds_file.clone();
        threads.push(std::thread::spawn(move || {
            winnable_seeds_thread(
                &setup2, 
                source,
                |seed| {
                    let mut f = wsf.lock().expect("must be able to lock");
                    write!(f, "{}\n", seed).expect("write should succeed");
                    f.flush().expect("flush should succeed");
                },
            );
        }));
    }

    for i in threads {
        let _ = i.join();
    }

}

#[derive(Clone)]
struct WorkSource {
    seed: Arc<Mutex<Option<u32>>>
}

impl WorkSource {
    fn new(seed: u32) -> Self {
        return WorkSource { seed: Arc::new(Mutex::new(Some(seed))) }
    }

    fn take(&self) -> Option<u32> {
        let mut self_seed = self.seed.lock().expect("shouldn't have panicked");
        if let Some(seed) = *self_seed {
            *self_seed = seed.checked_add(1);
            return Some(seed);
        }
        return None
    }
}

fn winnable_seeds_thread(setup: &ruleset::Setup, source: WorkSource, mut cb: impl FnMut(u32)) {
    let mut winnable = 0;
    let mut total = 0;
    
    while let Some(seed) = source.take() {
        let mut board = Board::new(&setup);
        board.deal(Deal::deal(&setup, seed));
        board.display();

        if is_winnable(board) {
            winnable += 1;
            cb(seed);
        }
        total += 1;
        println!("winnable: {}/{} ({}%)", winnable, total, (100.0 * winnable as f32)/(total as f32));
    }

    // println!("Legal moves: {:#?}", board.legal_moves());

}

fn is_winnable(mut board: Board<'_>) -> bool {
    return 
        explore(20, &mut board)  ||
        explore(200, &mut board)
    ;
}

fn explore(max_depth: usize, board: &mut Board<'_>) -> bool {
    let mut seen = Seen::new();
    explore2(max_depth, 0, board, &mut seen)
}

fn explore2(max_depth: usize, depth: usize, board: &mut Board<'_>, seen: &mut Seen) -> bool {
    if depth > max_depth {
        return false;
    }

    if seen.contains(board.zobrist_key()) {
        return false
    }
    seen.add(board.zobrist_key());

    if board.is_won() {
        board.display();
        return true
    }

    for m in board.legal_moves() {
        let hash_1 = board.zobrist_key();
        board.perform(m);
        // println!("try: {:X?} {:?}", board.zobrist_key(), m);
        if explore2(max_depth, depth + 1, board, seen) {
            return true;
        }
        // println!("undo: {:X?} {:?}", board.zobrist_key(), m);
        board.undo();
        let hash_2 = board.zobrist_key();
        assert_eq!(hash_1, hash_2)
    }
    return false;
}